/*
 * $Id: ngrep.c,v 1.26 1999/05/19 17:41:39 jpr5 Exp $
 *
 */

#include <stdlib.h>
#include <string.h>
#include <signal.h>

#ifdef LINUX
#include <getopt.h>
#endif

#if defined(BSD) || defined(SOLARIS)
#include <unistd.h>
#include <sys/types.h>
#include <ctype.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <net/if.h>
#endif

#ifdef LIBC5
#include <linux/if_ether.h>
#else 
#include <netinet/if_ether.h> 
#endif 

#include <netinet/ip.h> 
#include <netinet/tcp.h>
#include <netinet/udp.h>

#include <pcap.h> 
#include <net/bpf.h> 

#include "regex.h"
#include "ngrep.h"


int snaplen = 65535, promisc = 1, to = 1000;

char pc_err[PCAP_ERRBUF_SIZE], *re_err;
int quiet = 0;

int re_match_word = 0, re_ignore_case = 0;
struct re_pattern_buffer pattern;
char *regex, *filter;

struct bpf_program pcapfilter;
struct in_addr net, mask;
char *dev = NULL;
int ip_offset;
pcap_t *pd;


int main(int argc, char **argv) {
  char c;

  signal(SIGINT,dealloc);

  while ((c = getopt(argc, argv, "iwqhd:")) != EOF) {
    switch (c) {
    case 'd': 
      dev = optarg;
      break;
    case 'q':
      quiet++;
      break;
    case 'w':
      re_match_word++;
      break;
    case 'i':
      re_ignore_case++;
      break;
    case 'h':
      usage(0);
    default:
      usage(-1);
    }
  }

  if (!dev) 
    if (!(dev = pcap_lookupdev(pc_err))) {
      perror(pc_err);
      exit(-1);
    }
  if (!quiet) printf("dev: %s\n", dev);

  if (!(regex = argv[optind++])) {
    fprintf(stderr,"can't grok without a regex\n");
    exit(-1);
  } else {
    re_syntax_options = RE_SYNTAX_EGREP;

    if (re_match_word) {
      char *word_regex = malloc(strlen(regex) + strlen(WORD_REGEX));
      sprintf(word_regex,WORD_REGEX,regex,regex,regex);
      regex = word_regex;
    }

    if (re_ignore_case) {
      char *s;
      int i;
      
      s = regex;
      while (*s) {
	if (isalpha(*s)) 
	  *s = tolower(*s);
	s++;
      }

      pattern.translate = (char*)malloc(256);
      s = pattern.translate;
      for (i = 0; i < 256; i++) 
	s[i] = i;
      for (i = 'A'; i <= 'Z'; i++) 
	s[i] = i + 32;
    }

    (const)re_err = re_compile_pattern(regex,strlen(regex),&pattern);
    if (re_err) {
      perror(re_err);
      exit(-1);
    }

    if (!re_match_word) regex = NULL;
   
    pattern.fastmap = (char*)malloc(256);

    if (re_compile_fastmap(&pattern)) {
      perror("fastmap compile failed");
      exit(-1);
    }
  }

  if ((pd = pcap_open_live(dev, snaplen, promisc, to, pc_err)) == NULL) {
    perror(pc_err);
    exit(-1);
  }

  if (pcap_lookupnet(dev,&net.s_addr,&mask.s_addr, pc_err) == -1) {
    perror(pc_err);
    exit(-1);
  } else if (!quiet) {
    printf("net: %s\n",inet_ntoa(net)); 
    printf("mask: %s\n",inet_ntoa(mask)); 
  }

  if (argv[optind]) {
    filter = get_filter(&argv[optind]);
    if (!quiet) printf("filter: %s\n",filter);

    if (pcap_compile(pd,&pcapfilter,filter,0,mask.s_addr)) {
      pcap_perror(pd,"couldn't compile pcap filter");
      exit(-1);
    }
    
    if (pcap_setfilter(pd,&pcapfilter)) {
      pcap_perror(pd,"couldn't set pcap filter");
      exit(-1);
    }
  }

  switch(pcap_datalink(pd)) {
  case DLT_EN10MB:
  case DLT_IEEE802:
    ip_offset = ETHHDR_SIZE;
    break;

  case DLT_SLIP: 
    ip_offset = SLIPHDR_SIZE;
    break;
    
  case DLT_PPP:
    ip_offset = PPPHDR_SIZE;
    break;

  case DLT_RAW: 
    ip_offset = RAWHDR_SIZE;
    break;

  default:
    fprintf(stderr,"fatal: unsupported interface type\n");
    exit(-1);
  }
  
  while (pcap_loop(pd,0,(pcap_handler)process,0));
}	


void process(u_char *data1, struct pcap_pkthdr* h, u_char *p) {
  struct ip* ip_packet = (struct ip *)(p + ip_offset);
  char *data;
  int len;

  switch (ip_packet->ip_p) {
  case IPPROTO_IP:
  case IPPROTO_TCP: {
    struct tcphdr* tcp = (struct tcphdr *)(((char *)ip_packet) +
					   ip_packet->ip_hl*4); 

    if (!quiet) {
      printf("#");
      fflush(stdout);
    }

    data = ((char*)tcp) + tcp->th_off*4;
    len = h->len - ip_offset
                 - (ip_packet->ip_hl*4)         
                 - tcp->th_off*4;               
    
    if (len && match(&pattern,data,len)) { 
      printf("\nT ");
      printf("%s:%d -",inet_ntoa(ip_packet->ip_src),ntohs(tcp->th_sport));
      printf("> %s:%d",inet_ntoa(ip_packet->ip_dst),ntohs(tcp->th_dport));
      printf(" [%c%c%c%c%c%c]\n",
	     (tcp->th_flags&TH_ACK)?'A':' ',
	     (tcp->th_flags&TH_SYN)?'S':' ',
	     (tcp->th_flags&TH_RST)?'R':' ',
	     (tcp->th_flags&TH_FIN)?'F':' ',
	     (tcp->th_flags&TH_URG)?'U':' ',
	     (tcp->th_flags&TH_PUSH)?'P':' ');
      
      dump(data,len);
    }
  }
  return;

  case IPPROTO_UDP: {
    struct udphdr* udp = (struct udphdr *)(((char *)ip_packet) +
					   ip_packet->ip_hl*4); 
    if (!quiet) {
      printf("#"); 
      fflush(stdout);
    }

    data = ((char*)udp) + sizeof(struct udphdr);
#ifdef LIBC5
    len = ntohs(udp->len) - sizeof(struct udphdr);
#else
    len = ntohs(udp->uh_ulen) - sizeof(struct udphdr);
#endif

    if (len && match(&pattern,data,len)) { 
      printf("\nU ");

#ifdef LIBC5
      printf("%s:%d ->",inet_ntoa(ip_packet->ip_src),ntohs(udp->source));
      printf(" %s:%d\n",inet_ntoa(ip_packet->ip_dst),ntohs(udp->dest));
#else
      printf("%s:%d ->",inet_ntoa(ip_packet->ip_src),ntohs(udp->uh_sport));
      printf(" %s:%d\n",inet_ntoa(ip_packet->ip_dst),ntohs(udp->uh_dport));
#endif

      dump(data,len);
    }
  }
  return;

  }
}


void dump(char *data, int len) {  
  if (len) {
    char *str = data;
    int j, i = 0;
    while (i < len) {
      printf("  ");

      for (j=0; j<60; j++) {
	if (i+j < len) 
	  if (isprint(str[j])) 
	    printf("%c",str[j]);
	  else printf(".");
	else printf(" ");
      }

      str += 60;
      i += j;

      printf("\n");
    }
  }
}


int match(struct re_pattern_buffer *pattern, char *data, int len) {
  switch(re_search(pattern,data,len,0,len,0)) {
    case -2: 
      perror("she's dead, jim\n");
      exit(-1);
    case -1:
      return 0;
    default: 
      return 1;
  }
}


char *get_filter(char **argv) {
  char **arg = argv, *filter;
  char *from, *to;
  int len = 0;

  if (!*arg)
    return NULL;

  while (*arg) 
    len += strlen(*arg++) + 1;

  if (!(filter = (char*)malloc(len+1))) 
    return NULL;

  memset(filter,0,len);

  arg = argv;
  to = filter;

  while ((from = *arg++)) {
    while ((*to++ = *from++));
    *(to-1) = ' ';
  }

  return filter;
}


void dealloc(int sig) {
  if (regex) free(regex);
  if (filter) free(filter);
  if (pattern.translate) free(pattern.translate);
  if (pattern.fastmap) free(pattern.fastmap);

  printf("exit\n");
  exit(0);
}


void usage(int e) {
  printf("usage: ngrep <-hiwq> <-d dev> [regex] <pcap filter logic>\n");
  exit(e);
}
